/* $Workfile: OsaLinuxC.c $   $Revision: 19 $    $Modtime: 9/26/08 5:07p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc internal

  @module

  @normal Copyright (C) Odin TeleSystems Inc., 1998-2003, All Rights Reserved

*******************************************************************************/

#ifdef DAVINCI
// Davinci build
#ifdef asmlinkage
#undef asmlinkage
#endif

#ifndef asmlinkage
#define asmlinkage  
#endif
#include <asm/semaphore.h>
#endif

#include "OsaLinuxC.h"
#include "OsaLinuxRtl.h"
#include "OsaLinuxCpp.h" 
#include "OsaRtl.h"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/interrupt.h>  
#endif

static OsaUint32 g_LogLevel = 1; // log errors only

/*---------------------------------------------------------------------------*\  
 from OsaLinuxCppRtl.h OsaLinuxCppRtl.cpp
\*---------------------------------------------------------------------------*/

void OsaSpinLockImpl_Ctor(void **pSLock)
{
  if (sizeof(spinlock_t) > 0) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    /* FIXME: */
    *pSLock = kmalloc(sizeof(spinlock_t), GFP_KERNEL);
#else
    if(in_interrupt() || irqs_disabled())
     *pSLock = kmalloc(sizeof(spinlock_t), GFP_ATOMIC);
    else *pSLock = kmalloc(sizeof(spinlock_t), GFP_KERNEL);
#endif
    if (*pSLock) {
      spin_lock_init((spinlock_t *) *pSLock);
    } else {
      OSA_EPRINTF(("OsaLinux: Out of memory OsaSpinLockImpl_Ctor!\n"));
    }
  } else {
    spin_lock_init((spinlock_t *) *pSLock);
  }
}

void OsaSpinLockImpl_Dtor(void **pSLock)
{
  if (sizeof(spinlock_t) > 0) {
    if (*pSLock) {
      kfree(*pSLock);
      *pSLock = NULL;
    }
  }
}

void OsaSpinLockImpl_Lock(void *pSLock, __u32 *flags)
{
  __u32 lflags;

  spin_lock_irqsave((spinlock_t *) pSLock, lflags);
  *flags = lflags;
}

void OsaSpinLockImpl_Unlock(void *pSLock, __u32 savedFlags)
{
  spin_unlock_irqrestore((spinlock_t *) pSLock, (unsigned long) savedFlags);
}

void OsaEventImpl_Ctor(void **pWQHead)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)    
  if (sizeof(wait_queue) > 0) {
    *pWQHead = kmalloc(sizeof(wait_queue), GFP_KERNEL);
    if (*pWQHead) {
      init_waitqueue_head((wait_queue *) *pWQHead);
    } else {
      OSA_EPRINTF(("OsaLinux: Out of memory OsaEventImpl_Ctor!\n"));
    }
  } else {
    init_waitqueue_head((wait_queue *) *pWQHead);
  }
#else
  if (sizeof(wait_queue_head_t) > 0) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    /* FIXME: */
    *pWQHead = kmalloc(sizeof(wait_queue_head_t), GFP_KERNEL);
#else
    if(in_interrupt() || irqs_disabled())
     *pWQHead = kmalloc(sizeof(wait_queue_head_t), GFP_ATOMIC);
    else *pWQHead = kmalloc(sizeof(wait_queue_head_t), GFP_KERNEL);
#endif
    if (*pWQHead) {
      init_waitqueue_head((wait_queue_head_t *) *pWQHead);
    } else {
      OSA_EPRINTF(("OsaLinux: Out of memory OsaEventImpl_Ctor!\n"));
    }
  } else {
    init_waitqueue_head((wait_queue_head_t *) *pWQHead);
  }
#endif
}

void OsaEventImpl_Dtor(void **pWQHead)
{
  if (*pWQHead) {

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)    
    if (sizeof(wait_queue) > 0) {
      kfree(*pWQHead);
    }
#else
    if (sizeof(wait_queue_head_t) > 0) {
      kfree(*pWQHead);
    }
#endif

  }
  *pWQHead = NULL;
}


void OsaEventImpl_wake_up(void **pWQHead)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)    
  wake_up((wait_queue **) pWQHead);
#else
  wake_up((wait_queue_head_t *) *pWQHead);
#endif
}


void OsaSemaphoreImpl_Ctor(void **pSemaphore)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    /* FIXME: */
    *pSemaphore = kmalloc(sizeof(struct semaphore), GFP_KERNEL);
#else
    if(in_interrupt() || irqs_disabled())
     *pSemaphore = kmalloc(sizeof(struct semaphore), GFP_ATOMIC);
    else *pSemaphore = kmalloc(sizeof(struct semaphore), GFP_KERNEL);
#endif
  if (*pSemaphore) {
    sema_init((struct semaphore *) *pSemaphore, 0);     // initial count 0
  } else {
    OSA_EPRINTF(("OsaLinux: Out of memory OsaSemaphoreImpl_Ctor!\n"));
  }
}

void OsaSemaphoreImpl_Dtor(void **pSemaphore)
{
  if (*pSemaphore) {
    kfree(*pSemaphore);
    *pSemaphore = NULL;
  }
}

void OsaSemaphoreImpl_Release(void **pSemaphore)
{
  up((struct semaphore *) *pSemaphore);
}

void OsaSemaphoreImpl_WaitForSemaphore(void **pSemaphore)
{
  down((struct semaphore *) *pSemaphore);
}


/*---------------------------------------------------------------------------*\  
 from OsaLinux.h and OsaLinux.cpp
\*---------------------------------------------------------------------------*/
  
void OsaLinuxUserEventDevice_Ctor(void **pWQHead)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)    
  if (sizeof(wait_queue) > 0) {
    *pWQHead = kmalloc(sizeof(wait_queue), GFP_KERNEL);
    if (*pWQHead) {
      init_waitqueue_head((wait_queue *) *pWQHead);
    } else {
      OSA_EPRINTF(("OsaLinux: Out of memory OsaLinuxUserEventDevice_Ctor!\n"));
    }
  } else {
    init_waitqueue_head((wait_queue *) *pWQHead);
  }
#else
  if (sizeof(wait_queue_head_t) > 0) {
    *pWQHead = kmalloc(sizeof(wait_queue_head_t), GFP_KERNEL);
    if (*pWQHead) {
      init_waitqueue_head((wait_queue_head_t *) *pWQHead);
    } else {
      OSA_EPRINTF(("OsaLinux: Out of memory OsaLinuxUserEventDevice_Ctor!\n"));
    }
  } else {
    init_waitqueue_head((wait_queue_head_t *) *pWQHead);
  }
#endif
}

void OsaLinuxUserEventDevice_Dtor(void **pWQHead)
{
  if (*pWQHead) {

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)    
    if (sizeof(wait_queue) > 0) {
      kfree(*pWQHead);
    }
#else
    if (sizeof(wait_queue_head_t) > 0) {
      kfree(*pWQHead);
    }
#endif

  }
  *pWQHead = NULL;
}

void OsaLinuxUserEventDevice_wake_up(void **pWQHead)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)    
  wake_up((wait_queue **) pWQHead);
#else
  wake_up((wait_queue_head_t *) *pWQHead);
#endif
}

/*---------------------------------------------------------------------------*\  
 from OsaLinux.h and OsaLinux.cpp
\*---------------------------------------------------------------------------*/

int OsaLinux_pci_present()
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  return pci_present();
#else
#ifdef CONFIG_PCI
  return 1;
#else
  return 0;
#endif
#endif
}


void OsaLinux_pci_read_config_word(void *dev, int where, u16 *val)
{
  pci_read_config_word((struct pci_dev *)dev, where, val);
}


void OsaLinux_pci_read_config_dword(void *dev, int where, u32 *val)
{
  pci_read_config_dword((struct pci_dev *)dev, where, val);
}


void OsaLinux_pci_write_config_dword(void *dev, int where, u32 val)
{
  pci_write_config_dword((struct pci_dev *)dev, where, val);
}

void * OsaLinux_pci_dev_next(void *from)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
  if(from == NULL) {
    return pci_devices;
  }
  else {
    return ((pci_dev*)from)->next;
  }
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  if(from == NULL) {
    // Forward search of PCI device list
    return pci_dev_g(pci_devices.next);
  }
  else {
    if((void *) from == (void *) pci_dev_g(pci_devices.prev)) {
     // End of list, stop the search
     return NULL;
    }
    else {
     from = (void *) pci_dev_g(((struct pci_dev*)from)->global_list.next);
     return from;

    }
  }

#else
  return (void*)pci_find_device(PCI_ANY_ID, PCI_ANY_ID, from);
#endif
}


OsaUint16 OsaLinux_pci_dev_device(void *dev)
{
 return ((struct pci_dev*)dev)->device;
}

OsaUint16 OsaLinux_pci_dev_vendor(void *dev)
{
 return ((struct pci_dev*)dev)->vendor;
}

unsigned int OsaLinux_pci_dev_irq(void *dev)
{
 return ((struct pci_dev*)dev)->irq;
}

unsigned int OsaLinux_pci_dev_resource_start(void *dev, int id)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
 return ((struct pci_dev*)dev)->base_address[id];
#else
 return ((struct pci_dev*)dev)->resource[id].start;
#endif
}


/*---------------------------------------------------------------------------*\
  from OsaLinux.h and OsaLinux.cpp

  for RedHat 9.0; in RedHat 8.0, it compiles without modifying followings
\*---------------------------------------------------------------------------*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20) 
  #include <linux/module.h>  
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) 
void OsaLinux_MOD_INC_USE_COUNT(void)
{
   MOD_INC_USE_COUNT;
}


void OsaLinux_MOD_DEC_USE_COUNT(void)
{
   MOD_DEC_USE_COUNT;
}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) 
const void * Osalinux_inter_module_get(const char * sModule)
{
  return inter_module_get(sModule);
}


void OsaLinux_inter_module_register(const char * sName, const void * pFunName)
{
  inter_module_register(sName, THIS_MODULE, pFunName);
}


void OsaLinux_inter_module_unregister(const char * sName)
{
  inter_module_unregister(sName);
}
#endif

void OsaLinux_pci_dev_enable(void *d)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 struct pci_dev *dd = d;
 pci_enable_device(dd);
#else
 /* FIXME: */
#endif
}


OsaInt32 OsaGetKernelStackAvailable(OsaUint32 *stackAvailable)
{
#ifdef DAVINCI
  *stackAvailable = 0;

  return OSA_RES_NOT_SUPPORTED;
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
  static unsigned long lowestStackValue = (12*1024); // State with large number
  
  unsigned long esp;

  __asm__ __volatile__("andl %%esp,%0" :
     "=r" (esp) : "0" (THREAD_SIZE - 1));

  // Unused stack
  *stackAvailable = esp;

  if(lowestStackValue > esp) {
    lowestStackValue = esp;
    // OSA_DPRINTF will/could call OsaGetKernelStackAvailable()
    printk(KERN_INFO "**** OsaLinux: Current lowest stack value = %d\n", (int) lowestStackValue);
  }

  // During driver initialization 1000 is usually reached. Then
  // the stack usage drops well below 1000. Set the upper limit 
  // high enough so that only if the stack is 60% used.
  if(*stackAvailable < 1300) { 
    // OSA_DPRINTF will/could call OsaGetKernelStackAvailable()
    printk(KERN_INFO "\nOsaLinux: Stack is about to overflow, stack available size = %d\n", *stackAvailable);
		dump_stack(); 
    printk(KERN_INFO "OsaLinux: end of dump_stack\n");
  }
   
  return OSA_RES_SUCCESS;
#else
  *stackAvailable = 0;

  return OSA_RES_NOT_SUPPORTED;
#endif
#endif
}

/******************************************************************************
    void
    OsaSetLogErrorEvent(IN OsaUint32 nLogLevelFactor)

******************************************************************************/
void
OsaSetLogErrorEvent(
    IN OsaUint32 nLogLevelFactor
    )
{
  g_LogLevel = nLogLevelFactor;
}

/******************************************************************************
    OsaUint32
    OsaGetLogErrorEvent(void)

******************************************************************************/
OsaUint32
OsaGetLogErrorEvent(
    void
    )
{
  return(g_LogLevel);
}

/******************************************************************************
    OsaUint32
    OsaLogErrorEvent(
        IN OsaUint32 ErrorCode,
        IN OsaUint32 UniqueID ,
        IN OsaUint32 DeviceType,
        IN char      String1  ,
        IN OsaUint16 String2  
        )

    IN OsaUint32 ErrorCode  Error code
    IN OsaUint32 UniqueID   Unique error ID, used when there is more than one 
                            place reporting the same ErrorCode. It is usually 
                            the OTX Result Code.
    IN OsaUint32 DeviceType OtxPhysicalDeviceTypeE or 0 (OTX_DEVICE_UNKNOWN) 
    IN char      String1    Optional substitution string 1.
    IN OsaUint16 String1    Optional substitution string 2.

    return Otx_Result
******************************************************************************/
OsaUint32
OsaLogErrorEvent(
    IN OsaUint32 ErrorCode,
    IN OsaUint32 UniqueID,
    IN OsaUint32 DeviceType,
    IN char      *String1,
    IN OsaUint16 *String2)
{

#if DBG
  // Debug mode log additional messages.
#else
    // Only log events for error status "OTX_FAILURE" and informational events
  if(((UniqueID & 0x80000000) == 0) && (ErrorCode != OSA_INFORMATIONAL_EVENT_LOG)) {
    // Not an error
    return(UniqueID);
  }
#endif

  if(OsaGetCurrentExeLevel() == OSA_EXE_LEVEL_DEVICE) {
    // Interrupt handler for a device. Adding messages to log file could slow
    // processing of interrupt.
    return(UniqueID);
  }

  if(String1) {
    printk(KERN_INFO "Otx Error Event: %s, Result Code 0x%x\n", String1, UniqueID);
  }
  else {
    printk(KERN_INFO "Otx Error Event: No error string available, Result Code 0x%x\n", UniqueID);
  }

  return(UniqueID);
}
